#!/usr/bin/env python
"""amarok2rhythmbox.py
    Script to move lyrics and statistics from Amarok 1.4 to Rhythmbox 0.12.
    http://erlug.linux.it/~da/soft/amarok2rhythmbox/

    Copyright (C) 2009 Davide Alberani <da@erlug.linux.it>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import re
import glob
import urllib
import shutil
import hashlib
import logging
from htmlentitydefs import name2codepoint

import sqlalchemy
import BeautifulSoup
from PIL import Image

__version__ = VERSION = '0.2'

# ----------------------------------------------------------------------
# Things to configure.

# Every path is relative to the HOME directory.
HOME = os.environ.get('HOME')

# Path to the amarok's (1.4) collection.db file and covers.
AMAROK_COLLECTION = '%s/.kde/share/apps/amarok/collection.db' % HOME
AMAROK_COVERS_GLOB = '%s/.kde/share/apps/amarok/albumcovers/*/*' % HOME

# URI to access amarok's database - change this only if you're using MySQL with
# something like 'mysql://USER:PWD@localhost/AMAROK_DB'.
AMAROK_DB_URI = 'sqlite:///%s' % AMAROK_COLLECTION

# Directory used to store lyrics and covers for rhythmbox.
RHYTHMBOX_LYRICS = '%s/.cache/rhythmbox/lyrics/' % HOME
RHYTHMBOX_COVERS = '%s/.cache/rhythmbox/covers/' % HOME
# The XML file used by rhythmbox to store its data.
RHYTHMBOX_XML = '%s/.local/share/rhythmbox/rhythmdb.xml' % HOME

# It's assumed that this is the encoding used for lyrics and file names.
ENCODING = 'utf8'

# What we're asked to do.
DO_SANITY_CHECKS = True
DO_LYRICS = True
DO_STATS = True
DO_COVERS = True

# Shows only messages of this importance or higher.
LOG_LEVEL = logging.INFO

# Nothing to configure below - go away!
# ----------------------------------------------------------------------


# Set up the logging facility.
logging.basicConfig(level=LOG_LEVEL,
                    format='%(asctime)s %(levelname)s: %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')

# Things needed to convert HTML entities.
name2codepoint['apos'] = 39
name2codepoint['quote'] = name2codepoint['quot']
re_entities = re.compile(r'&(%s);' % '|'.join(name2codepoint.keys()))

def convert_entity(match):
    """Convert a HTML/XML (non-numeric) entity to a unicode char."""
    entity = match.group(1)
    int_value = name2codepoint[entity]
    return unichr(int_value)


# Connect to the database and read the db schema.
engine = sqlalchemy.create_engine(AMAROK_DB_URI)
meta = sqlalchemy.MetaData()
meta.reflect(bind=engine)


# Fetch some useful data.
artist_id2name = dict([(t['id'], t['name']) for t
                         in engine.execute(meta.tables['artist'].select())])
logging.debug('%d artists found' % len(artist_id2name))

album_id2title = dict([(t['id'], t['name']) for t
                         in engine.execute(meta.tables['album'].select())])
logging.debug('%d albums found' % len(album_id2title))

url2song_title = {}
url2artist_id = {}
url2album_id = {}
artist_album = {} # Use a dictionary to improve performances.
for song in engine.execute(meta.tables['tags'].select()):
    url = song['url']
    artist_id = song['artist']
    album_id = song['album']
    url2artist_id[url] = artist_id
    url2album_id[url] = album_id
    url2song_title[url] = song['title']
    # Collect tuples of (artist, album) IDs.
    a_a = (artist_id, album_id)
    if a_a not in artist_album:
        artist_album[a_a] = None

logging.debug('%d songs with statistics found' % len(url2song_title))

def convert_url2album(url):
    """Convert an url to an album title (or None)."""
    album_id = url2album_id.get(url)
    if album_id is None:
        return None
    return album_id2title.get(album_id)


def convert_url2artist(url):
    """Convert an url to an artist name (or None)."""
    artist_id = url2artist_id.get(url)
    if artist_id is None:
        return None
    return artist_id2name.get(artist_id)


def sanity_checks():
    """Perform a sanity check on various table, to detect problems with
    the encoding."""
    logging.info('starting sanity checks')
    for table in 'tags', 'lyrics', 'artist', 'album', 'statistics':
        sel = engine.execute(meta.tables[table].select())
        count = 1
        for item in sel:
            try:
                str(item)
            except Exception:
                _msg = 'wrong char in table "%s", at row "%d"; ' % (table,
                                                                    count+1)
                _msg += 'please try to fix it with sqlitebrowser.'
                logging.critical(_msg)
                raise
            count += 1
    logging.info('sanity checks done')


def song_lyrics():
    """Extract lyrics and save them in ".lyric" files."""
    logging.info('extracting lyrics')
    lyrics_table = meta.tables['lyrics']
    for song in engine.execute(lyrics_table.select()):
        url = song['url']
        lyrics = song['lyrics']
        if not (url and lyrics):
            continue
        # Parse the surrounding XML and extract the HTML stored inside,
        # converting escaped HTML entities.
        xml_soup = BeautifulSoup.BeautifulStoneSoup(lyrics)
        html_lyrics = re_entities.sub(convert_entity,
                                    unicode(xml_soup.lyrics.string))
        # Extract only the text from the HTML.
        html_soup = BeautifulSoup.BeautifulSoup(html_lyrics)
        lyrics_text = u''.join(html_soup.findAll(text=True)).encode(ENCODING)
        lyrics_text = lyrics_text.strip()
        artist_name = convert_url2artist(url)
        song_title = url2song_title.get(url)
        if not (lyrics_text and artist_name and song_title):
            logging.warning('ignored lyrics for %s' % url.encode(ENCODING,
                                                                 'replace'))
            continue
        artist_name = artist_name.encode(ENCODING, 'replace').lower()
        song_title = song_title.encode(ENCODING, 'replace').lower()
        artist_dir = os.path.join(RHYTHMBOX_LYRICS,
                                  artist_name.replace(os.sep, '_'))
        if not os.path.isdir(artist_dir):
            os.makedirs(artist_dir)
        lyrics_file = os.path.join(artist_dir,
                                '%s.lyric' % song_title.replace(os.sep, '_'))
        # Do not overide existing files.
        if not os.path.exists(lyrics_file):
            try:
                fd = open(lyrics_file, 'w')
                fd.write(lyrics_text)
            except (IOError, OSError), e:
                logging.error('unable to write "%s": %s' % (lyrics_file, e))
            finally:
                fd.close()
    logging.info('lyrics extracted')


def song_stats():
    """Extract information about songs and store them in the rhythmbox's XML
    file."""
    logging.info('extracting songs statistics')
    fd = open(RHYTHMBOX_XML)
    xml = fd.read()
    fd.close()
    logging.info('parsing rhythmbox XML file')
    rhythmbox_soup = BeautifulSoup.BeautifulStoneSoup(xml)
    location2tag = {}
    # Build a dictionary used to map urls to XML "entry" tags.
    for tag in rhythmbox_soup.findAll('entry', type="song"):
        location_text = tag.location.string
        if location_text:
            location2tag[location_text.replace('file://', '.', 1)] = tag
    logging.info('rhythmbox XML file parsed')
    logging.info('reading statistics from amarok')
    # Fetch every information from the amarok's database.
    url2info = dict([(t['url'], (t['rating'], t['playcounter'],
                   t['createdate'], t['accessdate'])) for t
                   in engine.execute(meta.tables['statistics'].select())])
    logging.info('statistics from amarok read')
    logging.info('modifying the rhythmobx XML file')
    for url, info in url2info.iteritems():
        escaped_url = urllib.quote(url.encode(ENCODING, 'replace'))
        # Un-escape some chars used in XML (" - %22, too?).
        escaped_url = escaped_url.replace('%27', "'")
        escaped_url = escaped_url.replace('%28', "(")
        escaped_url = escaped_url.replace('%29', ")")
        escaped_url = escaped_url.replace('%2C', ",")
        escaped_url = escaped_url.replace('%21', "!")
        escaped_url = escaped_url.replace('%2B', "+")
        escaped_url = escaped_url.replace('%26', "&amp;") # pure craziness.
        song_tag = location2tag.get(escaped_url)
        if song_tag is None:
            logging.warning('ignored stats for "%s"' %
                            url.encode(ENCODING, 'replace'))
            continue
        rating, play_counter, create_date, access_date = info
        # 'rating', 'play-count' and 'last-played' are updated only if
        # not already present.
        if rating and not song_tag.rating:
            div, mod = divmod(rating, 2)
            rating = str(div + (0, 1)[mod])
            new_tag = BeautifulSoup.Tag(rhythmbox_soup, 'rating')
            new_tag.insert(0, BeautifulSoup.NavigableString(rating))
            song_tag.insert(-1, new_tag)
        if play_counter and not song_tag.find('play-count'):
            new_tag = BeautifulSoup.Tag(rhythmbox_soup, 'play-count')
            new_tag.insert(0, BeautifulSoup.NavigableString(str(play_counter)))
            song_tag.insert(-1, new_tag)
        if access_date and not song_tag.find('last-played'):
            new_tag = BeautifulSoup.Tag(rhythmbox_soup, 'last-played')
            new_tag.insert(0, BeautifulSoup.NavigableString(str(access_date)))
            song_tag.insert(-1, new_tag)
        # 'create_date' is always updated, if the tag is here.
        if create_date and song_tag.find('first-seen'):
            song_tag.find('first-seen').string.replaceWith(str(create_date))
    fd = open(RHYTHMBOX_XML, 'w')
    new_xml = str(rhythmbox_soup)
    fd.write(new_xml)
    fd.close()
    logging.info('songs statistics saved')


def album_covers():
    """Copy album covers."""
    if not os.path.isdir(RHYTHMBOX_COVERS):
        os.makedirs(RHYTHMBOX_COVERS)
    md5sum2file = {}
    for img_path in glob.glob(AMAROK_COVERS_GLOB):
        if img_path.endswith('@shadow') or '/tagcover/' in img_path:
            continue
        img_file = os.path.basename(img_path)
        img_size = None
        s_f = img_file.split('@', 1)
        if len(s_f) == 2:
            img_file = s_f[1]
            img_size = int(s_f[0])
        if img_size is None:
            try:
                img_size = max(Image.open(img_path).size)
            except IOError:
                logging.warning('unable to guess size of "%s"' % img_path)
                continue
        if img_file not in md5sum2file:
            md5sum2file[img_file] = (img_path, img_size)
        else:
            if img_size > md5sum2file[img_file][1]:
                md5sum2file[img_file] = (img_path, img_size)
    for artist_id, album_id in artist_album:
        artist = artist_id2name.get(artist_id)
        album = album_id2title.get(album_id)
        if not (artist and album):
            continue
        artist = artist.encode(ENCODING, 'replace')
        album = album.encode(ENCODING, 'replace')
        md5sum = hashlib.md5(artist.lower() + album.lower()).hexdigest()
        if md5sum not in md5sum2file:
            logging.debug('ignored cover for "%s - %s"' % (artist, album))
            continue
        orig_path = md5sum2file[md5sum][0]
        # XXX: it's assumed that every cover is stored as a PNG, by Amarok.
        dest_file = '%s - %s.png' % (artist, album)
        dest_path = os.path.join(RHYTHMBOX_COVERS, dest_file.replace('/', ''))
        if not os.path.isfile(dest_path):
            try:
                shutil.copyfile(orig_path, dest_path)
            except IOError, e:
                logging.error('unable to copy cover from "%s" to "%s"' %
                              (orig_file, dest_path))


# Run, baby, run!
if DO_SANITY_CHECKS:
    sanity_checks()
if DO_LYRICS:
    song_lyrics()
if DO_STATS:
    song_stats()
if DO_COVERS:
    album_covers()
